package com.sun.org.apache.xerces.internal.parsers;

import com.sun.org.apache.xerces.internal.dom.ElementDefinitionImpl;
import com.sun.org.apache.xerces.internal.dom.NotationImpl;
import com.sun.org.apache.xerces.internal.dom.DocumentTypeImpl;
import com.sun.org.apache.xerces.internal.dom.DOMErrorImpl;
import com.sun.org.apache.xerces.internal.dom.NodeImpl;
import org.w3c.dom.NodeList;
import com.sun.org.apache.xerces.internal.dom.PSVIElementNSImpl;
import com.sun.org.apache.xerces.internal.dom.TextImpl;
import org.w3c.dom.Text;
import com.sun.org.apache.xerces.internal.xs.XSTypeDefinition;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import com.sun.org.apache.xerces.internal.dom.ElementNSImpl;
import com.sun.org.apache.xerces.internal.xs.ElementPSVI;
import com.sun.org.apache.xerces.internal.dom.ElementImpl;
import com.sun.org.apache.xerces.internal.dom.AttrImpl;
import com.sun.org.apache.xerces.internal.dom.PSVIAttrNSImpl;
import com.sun.org.apache.xerces.internal.impl.dv.XSSimpleType;
import com.sun.org.apache.xerces.internal.xs.AttributePSVI;
import com.sun.org.apache.xerces.internal.xni.XMLAttributes;
import com.sun.org.apache.xerces.internal.dom.DocumentImpl;
import com.sun.org.apache.xerces.internal.dom.PSVIDocumentImpl;
import com.sun.org.apache.xerces.internal.xni.NamespaceContext;
import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.Comment;
import com.sun.org.apache.xerces.internal.xni.XMLString;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.EntityReference;
import com.sun.org.apache.xerces.internal.dom.EntityReferenceImpl;
import com.sun.org.apache.xerces.internal.xni.Augmentations;
import com.sun.org.apache.xerces.internal.xni.XMLResourceIdentifier;
import java.util.Locale;
import com.sun.org.apache.xerces.internal.xni.XNIException;
import com.sun.org.apache.xerces.internal.dom.DOMMessageFormatter;
import com.sun.org.apache.xerces.internal.utils.ObjectFactory;
import com.sun.org.apache.xerces.internal.xni.parser.XMLParserConfiguration;
import org.w3c.dom.ls.LSParserFilter;
import com.sun.org.apache.xerces.internal.xni.XMLLocator;
import com.sun.org.apache.xerces.internal.xni.QName;
import java.util.Stack;
import com.sun.org.apache.xerces.internal.dom.DeferredDocumentImpl;
import com.sun.org.apache.xerces.internal.dom.EntityImpl;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Node;
import org.w3c.dom.DocumentType;
import com.sun.org.apache.xerces.internal.dom.CoreDocumentImpl;
import org.w3c.dom.Document;
import com.sun.org.apache.xerces.internal.util.DOMErrorHandlerWrapper;

public class AbstractDOMParser extends AbstractXMLDocumentParser
{
    protected static final String NAMESPACES = "http://xml.org/sax/features/namespaces";
    protected static final String CREATE_ENTITY_REF_NODES = "http://apache.org/xml/features/dom/create-entity-ref-nodes";
    protected static final String INCLUDE_COMMENTS_FEATURE = "http://apache.org/xml/features/include-comments";
    protected static final String CREATE_CDATA_NODES_FEATURE = "http://apache.org/xml/features/create-cdata-nodes";
    protected static final String INCLUDE_IGNORABLE_WHITESPACE = "http://apache.org/xml/features/dom/include-ignorable-whitespace";
    protected static final String DEFER_NODE_EXPANSION = "http://apache.org/xml/features/dom/defer-node-expansion";
    private static final String[] RECOGNIZED_FEATURES;
    protected static final String DOCUMENT_CLASS_NAME = "http://apache.org/xml/properties/dom/document-class-name";
    protected static final String CURRENT_ELEMENT_NODE = "http://apache.org/xml/properties/dom/current-element-node";
    private static final String[] RECOGNIZED_PROPERTIES;
    protected static final String DEFAULT_DOCUMENT_CLASS_NAME = "com.sun.org.apache.xerces.internal.dom.DocumentImpl";
    protected static final String CORE_DOCUMENT_CLASS_NAME = "com.sun.org.apache.xerces.internal.dom.CoreDocumentImpl";
    protected static final String PSVI_DOCUMENT_CLASS_NAME = "com.sun.org.apache.xerces.internal.dom.PSVIDocumentImpl";
    private static final boolean DEBUG_EVENTS = false;
    private static final boolean DEBUG_BASEURI = false;
    protected DOMErrorHandlerWrapper fErrorHandler;
    protected boolean fInDTD;
    protected boolean fCreateEntityRefNodes;
    protected boolean fIncludeIgnorableWhitespace;
    protected boolean fIncludeComments;
    protected boolean fCreateCDATANodes;
    protected Document fDocument;
    protected CoreDocumentImpl fDocumentImpl;
    protected boolean fStorePSVI;
    protected String fDocumentClassName;
    protected DocumentType fDocumentType;
    protected Node fCurrentNode;
    protected CDATASection fCurrentCDATASection;
    protected EntityImpl fCurrentEntityDecl;
    protected int fDeferredEntityDecl;
    protected final StringBuilder fStringBuilder;
    protected StringBuilder fInternalSubset;
    protected boolean fDeferNodeExpansion;
    protected boolean fNamespaceAware;
    protected DeferredDocumentImpl fDeferredDocumentImpl;
    protected int fDocumentIndex;
    protected int fDocumentTypeIndex;
    protected int fCurrentNodeIndex;
    protected int fCurrentCDATASectionIndex;
    protected boolean fInDTDExternalSubset;
    protected Node fRoot;
    protected boolean fInCDATASection;
    protected boolean fFirstChunk;
    protected boolean fFilterReject;
    protected final Stack fBaseURIStack;
    protected int fRejectedElementDepth;
    protected Stack fSkippedElemStack;
    protected boolean fInEntityRef;
    private final QName fAttrQName;
    private XMLLocator fLocator;
    protected LSParserFilter fDOMFilter;
    
    protected AbstractDOMParser(XMLParserConfiguration config) {
        super(config);
        this.fErrorHandler = null;
        this.fStringBuilder = new StringBuilder(50);
        this.fFirstChunk = false;
        this.fFilterReject = false;
        this.fBaseURIStack = new Stack();
        this.fRejectedElementDepth = 0;
        this.fSkippedElemStack = null;
        this.fInEntityRef = false;
        this.fAttrQName = new QName();
        this.fDOMFilter = null;
        this.fConfiguration.addRecognizedFeatures(AbstractDOMParser.RECOGNIZED_FEATURES);
        this.fConfiguration.setFeature("http://apache.org/xml/features/dom/create-entity-ref-nodes", true);
        this.fConfiguration.setFeature("http://apache.org/xml/features/dom/include-ignorable-whitespace", true);
        this.fConfiguration.setFeature("http://apache.org/xml/features/dom/defer-node-expansion", true);
        this.fConfiguration.setFeature("http://apache.org/xml/features/include-comments", true);
        this.fConfiguration.setFeature("http://apache.org/xml/features/create-cdata-nodes", true);
        this.fConfiguration.addRecognizedProperties(AbstractDOMParser.RECOGNIZED_PROPERTIES);
        this.fConfiguration.setProperty("http://apache.org/xml/properties/dom/document-class-name", "com.sun.org.apache.xerces.internal.dom.DocumentImpl");
    }
    
    protected String getDocumentClassName() {
        return this.fDocumentClassName;
    }
    
    protected void setDocumentClassName(String documentClassName) {
        if (documentClassName == null) {
            documentClassName = "com.sun.org.apache.xerces.internal.dom.DocumentImpl";
        }
        if (!documentClassName.equals("com.sun.org.apache.xerces.internal.dom.DocumentImpl") && !documentClassName.equals("com.sun.org.apache.xerces.internal.dom.PSVIDocumentImpl")) {
            try {
                Class<Object> _class = ObjectFactory.findProviderClass(documentClassName, true);
                if (!Document.class.isAssignableFrom(_class)) {
                    throw new IllegalArgumentException(DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "InvalidDocumentClassName", new Object[] { documentClassName }));
                }
            }
            catch (ClassNotFoundException e) {
                throw new IllegalArgumentException(DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "MissingDocumentClassName", new Object[] { documentClassName }));
            }
        }
        this.fDocumentClassName = documentClassName;
        if (!documentClassName.equals("com.sun.org.apache.xerces.internal.dom.DocumentImpl")) {
            this.fDeferNodeExpansion = false;
        }
    }
    
    public Document getDocument() {
        return this.fDocument;
    }
    
    public final void dropDocumentReferences() {
        this.fDocument = null;
        this.fDocumentImpl = null;
        this.fDeferredDocumentImpl = null;
        this.fDocumentType = null;
        this.fCurrentNode = null;
        this.fCurrentCDATASection = null;
        this.fCurrentEntityDecl = null;
        this.fRoot = null;
    }
    
    public void reset() throws XNIException {
        super.reset();
        this.fCreateEntityRefNodes = this.fConfiguration.getFeature("http://apache.org/xml/features/dom/create-entity-ref-nodes");
        this.fIncludeIgnorableWhitespace = this.fConfiguration.getFeature("http://apache.org/xml/features/dom/include-ignorable-whitespace");
        this.fDeferNodeExpansion = this.fConfiguration.getFeature("http://apache.org/xml/features/dom/defer-node-expansion");
        this.fNamespaceAware = this.fConfiguration.getFeature("http://xml.org/sax/features/namespaces");
        this.fIncludeComments = this.fConfiguration.getFeature("http://apache.org/xml/features/include-comments");
        this.fCreateCDATANodes = this.fConfiguration.getFeature("http://apache.org/xml/features/create-cdata-nodes");
        this.setDocumentClassName((String)this.fConfiguration.getProperty("http://apache.org/xml/properties/dom/document-class-name"));
        this.fDocument = null;
        this.fDocumentImpl = null;
        this.fStorePSVI = false;
        this.fDocumentType = null;
        this.fDocumentTypeIndex = -1;
        this.fDeferredDocumentImpl = null;
        this.fCurrentNode = null;
        this.fStringBuilder.setLength(0);
        this.fRoot = null;
        this.fInDTD = false;
        this.fInDTDExternalSubset = false;
        this.fInCDATASection = false;
        this.fFirstChunk = false;
        this.fCurrentCDATASection = null;
        this.fCurrentCDATASectionIndex = -1;
        this.fBaseURIStack.removeAllElements();
    }
    
    public void setLocale(Locale locale) {
        this.fConfiguration.setLocale(locale);
    }
    
    @Override
    public void startGeneralEntity(String name, XMLResourceIdentifier identifier, String encoding, Augmentations augs) throws XNIException {
        if (!this.fDeferNodeExpansion) {
            if (this.fFilterReject) {
                return;
            }
            this.setCharacterData(true);
            EntityReference er2 = this.fDocument.createEntityReference(name);
            if (this.fDocumentImpl != null) {
                EntityReferenceImpl erImpl = (EntityReferenceImpl)er2;
                erImpl.setBaseURI(identifier.getExpandedSystemId());
                if (this.fDocumentType != null) {
                    NamedNodeMap entities = this.fDocumentType.getEntities();
                    this.fCurrentEntityDecl = (EntityImpl)entities.getNamedItem(name);
                    if (this.fCurrentEntityDecl != null) {
                        this.fCurrentEntityDecl.setInputEncoding(encoding);
                    }
                }
                erImpl.needsSyncChildren(false);
            }
            this.fInEntityRef = true;
            this.fCurrentNode.appendChild(er2);
            this.fCurrentNode = er2;
        }
        else {
            int er = this.fDeferredDocumentImpl.createDeferredEntityReference(name, identifier.getExpandedSystemId());
            if (this.fDocumentTypeIndex != -1) {
                for (int node = this.fDeferredDocumentImpl.getLastChild(this.fDocumentTypeIndex, false); node != -1; node = this.fDeferredDocumentImpl.getRealPrevSibling(node, false)) {
                    short nodeType = this.fDeferredDocumentImpl.getNodeType(node, false);
                    if (nodeType == 6) {
                        String nodeName = this.fDeferredDocumentImpl.getNodeName(node, false);
                        if (nodeName.equals(name)) {
                            this.fDeferredEntityDecl = node;
                            this.fDeferredDocumentImpl.setInputEncoding(node, encoding);
                            break;
                        }
                    }
                }
            }
            this.fDeferredDocumentImpl.appendChild(this.fCurrentNodeIndex, er);
            this.fCurrentNodeIndex = er;
        }
    }
    
    @Override
    public void textDecl(String version, String encoding, Augmentations augs) throws XNIException {
        if (this.fInDTD) {
            return;
        }
        if (!this.fDeferNodeExpansion) {
            if (this.fCurrentEntityDecl != null && !this.fFilterReject) {
                this.fCurrentEntityDecl.setXmlEncoding(encoding);
                if (version != null) {
                    this.fCurrentEntityDecl.setXmlVersion(version);
                }
            }
        }
        else if (this.fDeferredEntityDecl != -1) {
            this.fDeferredDocumentImpl.setEntityInfo(this.fDeferredEntityDecl, version, encoding);
        }
        return;
    }
    
    @Override
    public void comment(XMLString text, Augmentations augs) throws XNIException {
        if (this.fInDTD) {
            if (this.fInternalSubset != null && !this.fInDTDExternalSubset) {
                this.fInternalSubset.append("<!--");
                if (text.length > 0) {
                    this.fInternalSubset.append(text.ch, text.offset, text.length);
                }
                this.fInternalSubset.append("-->");
            }
            return;
        }
        else {
            if (!this.fIncludeComments || this.fFilterReject) {
                return;
            }
            if (!this.fDeferNodeExpansion) {
                Comment comment2 = this.fDocument.createComment(text.toString());
                this.setCharacterData(false);
                this.fCurrentNode.appendChild(comment2);
                if (this.fDOMFilter != null && !this.fInEntityRef && (this.fDOMFilter.getWhatToShow() & 0x80) != 0x0) {
                    short code = this.fDOMFilter.acceptNode(comment2);
                    switch (code) {
                        case 4: {
                            throw AbstractDOMParser.Abort.INSTANCE;
                        }
                        case 2:
                        case 3: {
                            this.fCurrentNode.removeChild(comment2);
                            this.fFirstChunk = true;
                            return;
                        }
                        default: {
                            break;
                        }
                    }
                }
            }
            else {
                int comment = this.fDeferredDocumentImpl.createDeferredComment(text.toString());
                this.fDeferredDocumentImpl.appendChild(this.fCurrentNodeIndex, comment);
            }
            return;
        }
    }
    
    @Override
    public void processingInstruction(String target, XMLString data, Augmentations augs) throws XNIException {
        if (this.fInDTD) {
            if (this.fInternalSubset != null && !this.fInDTDExternalSubset) {
                this.fInternalSubset.append("<?");
                this.fInternalSubset.append(target);
                if (data.length > 0) {
                    this.fInternalSubset.append(' ').append(data.ch, data.offset, data.length);
                }
                this.fInternalSubset.append("?>");
            }
            return;
        }
        else {
            if (!this.fDeferNodeExpansion) {
                if (this.fFilterReject) {
                    return;
                }
                ProcessingInstruction pi2 = this.fDocument.createProcessingInstruction(target, data.toString());
                this.setCharacterData(false);
                this.fCurrentNode.appendChild(pi2);
                if (this.fDOMFilter != null && !this.fInEntityRef && (this.fDOMFilter.getWhatToShow() & 0x40) != 0x0) {
                    short code = this.fDOMFilter.acceptNode(pi2);
                    switch (code) {
                        case 4: {
                            throw AbstractDOMParser.Abort.INSTANCE;
                        }
                        case 2:
                        case 3: {
                            this.fCurrentNode.removeChild(pi2);
                            this.fFirstChunk = true;
                            return;
                        }
                        default: {
                            break;
                        }
                    }
                }
            }
            else {
                int pi = this.fDeferredDocumentImpl.createDeferredProcessingInstruction(target, data.toString());
                this.fDeferredDocumentImpl.appendChild(this.fCurrentNodeIndex, pi);
            }
            return;
        }
    }
    
    @Override
    public void startDocument(XMLLocator locator, String encoding, NamespaceContext namespaceContext, Augmentations augs) throws XNIException {
        this.fLocator = locator;
        if (!this.fDeferNodeExpansion) {
            if (this.fDocumentClassName.equals("com.sun.org.apache.xerces.internal.dom.DocumentImpl")) {
                this.fDocument = new DocumentImpl();
                (this.fDocumentImpl = (CoreDocumentImpl)this.fDocument).setStrictErrorChecking(false);
                this.fDocumentImpl.setInputEncoding(encoding);
                this.fDocumentImpl.setDocumentURI(locator.getExpandedSystemId());
            }
            else if (this.fDocumentClassName.equals("com.sun.org.apache.xerces.internal.dom.PSVIDocumentImpl")) {
                this.fDocument = new PSVIDocumentImpl();
                this.fDocumentImpl = (CoreDocumentImpl)this.fDocument;
                this.fStorePSVI = true;
                this.fDocumentImpl.setStrictErrorChecking(false);
                this.fDocumentImpl.setInputEncoding(encoding);
                this.fDocumentImpl.setDocumentURI(locator.getExpandedSystemId());
            }
            else {
                try {
                    Class<Object> documentClass = ObjectFactory.findProviderClass(this.fDocumentClassName, true);
                    this.fDocument = (Document)documentClass.getConstructor(new Class[0]).newInstance(new Object[0]);
                    Class<Object> defaultDocClass = ObjectFactory.findProviderClass("com.sun.org.apache.xerces.internal.dom.CoreDocumentImpl", true);
                    if (defaultDocClass.isAssignableFrom(documentClass)) {
                        this.fDocumentImpl = (CoreDocumentImpl)this.fDocument;
                        Class<Object> psviDocClass = ObjectFactory.findProviderClass("com.sun.org.apache.xerces.internal.dom.PSVIDocumentImpl", true);
                        if (psviDocClass.isAssignableFrom(documentClass)) {
                            this.fStorePSVI = true;
                        }
                        this.fDocumentImpl.setStrictErrorChecking(false);
                        this.fDocumentImpl.setInputEncoding(encoding);
                        if (locator != null) {
                            this.fDocumentImpl.setDocumentURI(locator.getExpandedSystemId());
                        }
                    }
                }
                catch (ClassNotFoundException ex) {}
                catch (Exception e) {
                    throw new RuntimeException(DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "CannotCreateDocumentClass", new Object[] { this.fDocumentClassName }));
                }
            }
            this.fCurrentNode = this.fDocument;
        }
        else {
            this.fDeferredDocumentImpl = new DeferredDocumentImpl(this.fNamespaceAware);
            this.fDocument = this.fDeferredDocumentImpl;
            this.fDocumentIndex = this.fDeferredDocumentImpl.createDeferredDocument();
            this.fDeferredDocumentImpl.setInputEncoding(encoding);
            this.fDeferredDocumentImpl.setDocumentURI(locator.getExpandedSystemId());
            this.fCurrentNodeIndex = this.fDocumentIndex;
        }
    }
    
    @Override
    public void xmlDecl(String version, String encoding, String standalone, Augmentations augs) throws XNIException {
        if (!this.fDeferNodeExpansion) {
            if (this.fDocumentImpl != null) {
                if (version != null) {
                    this.fDocumentImpl.setXmlVersion(version);
                }
                this.fDocumentImpl.setXmlEncoding(encoding);
                this.fDocumentImpl.setXmlStandalone("yes".equals(standalone));
            }
        }
        else {
            if (version != null) {
                this.fDeferredDocumentImpl.setXmlVersion(version);
            }
            this.fDeferredDocumentImpl.setXmlEncoding(encoding);
            this.fDeferredDocumentImpl.setXmlStandalone("yes".equals(standalone));
        }
    }
    
    @Override
    public void doctypeDecl(String rootElement, String publicId, String systemId, Augmentations augs) throws XNIException {
        if (!this.fDeferNodeExpansion) {
            if (this.fDocumentImpl != null) {
                this.fDocumentType = this.fDocumentImpl.createDocumentType(rootElement, publicId, systemId);
                this.fCurrentNode.appendChild(this.fDocumentType);
            }
        }
        else {
            this.fDocumentTypeIndex = this.fDeferredDocumentImpl.createDeferredDocumentType(rootElement, publicId, systemId);
            this.fDeferredDocumentImpl.appendChild(this.fCurrentNodeIndex, this.fDocumentTypeIndex);
        }
    }
    
    @Override
    public void startElement(QName element, XMLAttributes attributes, Augmentations augs) throws XNIException {
        if (!this.fDeferNodeExpansion) {
            if (this.fFilterReject) {
                ++this.fRejectedElementDepth;
                return;
            }
            Element el2 = this.createElementNode(element);
            int attrCount2 = attributes.getLength();
            boolean seenSchemaDefault = false;
            for (int i = 0; i < attrCount2; ++i) {
                attributes.getName(i, this.fAttrQName);
                Attr attr = this.createAttrNode(this.fAttrQName);
                String attrValue = attributes.getValue(i);
                AttributePSVI attrPSVI2 = (AttributePSVI)attributes.getAugmentations(i).getItem("ATTRIBUTE_PSVI");
                if (this.fStorePSVI && attrPSVI2 != null) {
                    ((PSVIAttrNSImpl)attr).setPSVI(attrPSVI2);
                }
                attr.setValue(attrValue);
                boolean specified = attributes.isSpecified(i);
                if (!specified && (seenSchemaDefault || (this.fAttrQName.uri != null && this.fAttrQName.uri != NamespaceContext.XMLNS_URI && this.fAttrQName.prefix == null))) {
                    el2.setAttributeNodeNS(attr);
                    seenSchemaDefault = true;
                }
                else {
                    el2.setAttributeNode(attr);
                }
                if (this.fDocumentImpl != null) {
                    AttrImpl attrImpl = (AttrImpl)attr;
                    Object type2 = null;
                    boolean id2 = false;
                    if (attrPSVI2 != null && this.fNamespaceAware) {
                        type2 = attrPSVI2.getMemberTypeDefinition();
                        if (type2 == null) {
                            type2 = attrPSVI2.getTypeDefinition();
                            if (type2 != null) {
                                id2 = ((XSSimpleType)type2).isIDType();
                                attrImpl.setType(type2);
                            }
                        }
                        else {
                            id2 = ((XSSimpleType)type2).isIDType();
                            attrImpl.setType(type2);
                        }
                    }
                    else {
                        boolean isDeclared2 = Boolean.TRUE.equals(attributes.getAugmentations(i).getItem("ATTRIBUTE_DECLARED"));
                        if (isDeclared2) {
                            type2 = attributes.getType(i);
                            id2 = "ID".equals(type2);
                        }
                        attrImpl.setType(type2);
                    }
                    if (id2) {
                        ((ElementImpl)el2).setIdAttributeNode(attr, true);
                    }
                    attrImpl.setSpecified(specified);
                }
            }
            this.setCharacterData(false);
            if (augs != null) {
                ElementPSVI elementPSVI = (ElementPSVI)augs.getItem("ELEMENT_PSVI");
                if (elementPSVI != null && this.fNamespaceAware) {
                    XSTypeDefinition type3 = elementPSVI.getMemberTypeDefinition();
                    if (type3 == null) {
                        type3 = elementPSVI.getTypeDefinition();
                    }
                    ((ElementNSImpl)el2).setType(type3);
                }
            }
            if (this.fDOMFilter != null && !this.fInEntityRef) {
                if (this.fRoot == null) {
                    this.fRoot = el2;
                }
                else {
                    short code = this.fDOMFilter.startElement(el2);
                    switch (code) {
                        case 4: {
                            throw AbstractDOMParser.Abort.INSTANCE;
                        }
                        case 2: {
                            this.fFilterReject = true;
                            this.fRejectedElementDepth = 0;
                            return;
                        }
                        case 3: {
                            this.fFirstChunk = true;
                            this.fSkippedElemStack.push(Boolean.TRUE);
                            return;
                        }
                        default: {
                            if (!this.fSkippedElemStack.isEmpty()) {
                                this.fSkippedElemStack.push(Boolean.FALSE);
                                break;
                            }
                            break;
                        }
                    }
                }
            }
            this.fCurrentNode.appendChild(el2);
            this.fCurrentNode = el2;
        }
        else {
            int el = this.fDeferredDocumentImpl.createDeferredElement(this.fNamespaceAware ? element.uri : null, element.rawname);
            Object type = null;
            int attrCount = attributes.getLength();
            for (int i = attrCount - 1; i >= 0; --i) {
                AttributePSVI attrPSVI = (AttributePSVI)attributes.getAugmentations(i).getItem("ATTRIBUTE_PSVI");
                boolean id = false;
                if (attrPSVI != null && this.fNamespaceAware) {
                    type = attrPSVI.getMemberTypeDefinition();
                    if (type == null) {
                        type = attrPSVI.getTypeDefinition();
                        if (type != null) {
                            id = ((XSSimpleType)type).isIDType();
                        }
                    }
                    else {
                        id = ((XSSimpleType)type).isIDType();
                    }
                }
                else {
                    boolean isDeclared = Boolean.TRUE.equals(attributes.getAugmentations(i).getItem("ATTRIBUTE_DECLARED"));
                    if (isDeclared) {
                        type = attributes.getType(i);
                        id = "ID".equals(type);
                    }
                }
                this.fDeferredDocumentImpl.setDeferredAttribute(el, attributes.getQName(i), attributes.getURI(i), attributes.getValue(i), attributes.isSpecified(i), id, type);
            }
            this.fDeferredDocumentImpl.appendChild(this.fCurrentNodeIndex, el);
            this.fCurrentNodeIndex = el;
        }
    }
    
    @Override
    public void emptyElement(QName element, XMLAttributes attributes, Augmentations augs) throws XNIException {
        this.startElement(element, attributes, augs);
        this.endElement(element, augs);
    }
    
    @Override
    public void characters(XMLString text, Augmentations augs) throws XNIException {
        if (!this.fDeferNodeExpansion) {
            if (this.fFilterReject) {
                return;
            }
            if (this.fInCDATASection && this.fCreateCDATANodes) {
                if (this.fCurrentCDATASection == null) {
                    this.fCurrentCDATASection = this.fDocument.createCDATASection(text.toString());
                    this.fCurrentNode.appendChild(this.fCurrentCDATASection);
                    this.fCurrentNode = this.fCurrentCDATASection;
                }
                else {
                    this.fCurrentCDATASection.appendData(text.toString());
                }
            }
            else if (!this.fInDTD) {
                if (text.length == 0) {
                    return;
                }
                Node child = this.fCurrentNode.getLastChild();
                if (child != null && child.getNodeType() == 3) {
                    if (this.fFirstChunk) {
                        if (this.fDocumentImpl != null) {
                            this.fStringBuilder.append(((TextImpl)child).removeData());
                        }
                        else {
                            this.fStringBuilder.append(((Text)child).getData());
                            ((Node)child).setNodeValue(null);
                        }
                        this.fFirstChunk = false;
                    }
                    if (text.length > 0) {
                        this.fStringBuilder.append(text.ch, text.offset, text.length);
                    }
                }
                else {
                    this.fFirstChunk = true;
                    Text textNode = this.fDocument.createTextNode(text.toString());
                    this.fCurrentNode.appendChild(textNode);
                }
            }
        }
        else if (this.fInCDATASection && this.fCreateCDATANodes) {
            if (this.fCurrentCDATASectionIndex == -1) {
                int cs = this.fDeferredDocumentImpl.createDeferredCDATASection(text.toString());
                this.fDeferredDocumentImpl.appendChild(this.fCurrentNodeIndex, cs);
                this.fCurrentCDATASectionIndex = cs;
                this.fCurrentNodeIndex = cs;
            }
            else {
                int txt2 = this.fDeferredDocumentImpl.createDeferredTextNode(text.toString(), false);
                this.fDeferredDocumentImpl.appendChild(this.fCurrentNodeIndex, txt2);
            }
        }
        else if (!this.fInDTD) {
            if (text.length == 0) {
                return;
            }
            String value = text.toString();
            int txt = this.fDeferredDocumentImpl.createDeferredTextNode(value, false);
            this.fDeferredDocumentImpl.appendChild(this.fCurrentNodeIndex, txt);
        }
    }
    
    @Override
    public void ignorableWhitespace(XMLString text, Augmentations augs) throws XNIException {
        if (!this.fIncludeIgnorableWhitespace || this.fFilterReject) {
            return;
        }
        if (!this.fDeferNodeExpansion) {
            Node child = this.fCurrentNode.getLastChild();
            if (child != null && child.getNodeType() == 3) {
                Text textNode = (Text)child;
                textNode.appendData(text.toString());
            }
            else {
                Text textNode = this.fDocument.createTextNode(text.toString());
                if (this.fDocumentImpl != null) {
                    TextImpl textNodeImpl = (TextImpl)textNode;
                    textNodeImpl.setIgnorableWhitespace(true);
                }
                this.fCurrentNode.appendChild(textNode);
            }
        }
        else {
            int txt = this.fDeferredDocumentImpl.createDeferredTextNode(text.toString(), true);
            this.fDeferredDocumentImpl.appendChild(this.fCurrentNodeIndex, txt);
        }
        return;
    }
    
    @Override
    public void endElement(QName element, Augmentations augs) throws XNIException {
        if (!this.fDeferNodeExpansion) {
            if (augs != null && this.fDocumentImpl != null && (this.fNamespaceAware || this.fStorePSVI)) {
                ElementPSVI elementPSVI = (ElementPSVI)augs.getItem("ELEMENT_PSVI");
                if (elementPSVI != null) {
                    if (this.fNamespaceAware) {
                        XSTypeDefinition type = elementPSVI.getMemberTypeDefinition();
                        if (type == null) {
                            type = elementPSVI.getTypeDefinition();
                        }
                        ((ElementNSImpl)this.fCurrentNode).setType(type);
                    }
                    if (this.fStorePSVI) {
                        ((PSVIElementNSImpl)this.fCurrentNode).setPSVI(elementPSVI);
                    }
                }
            }
            if (this.fDOMFilter != null) {
                if (this.fFilterReject) {
                    if (this.fRejectedElementDepth-- == 0) {
                        this.fFilterReject = false;
                    }
                    return;
                }
                else {
                    if (!this.fSkippedElemStack.isEmpty() && this.fSkippedElemStack.pop() == Boolean.TRUE) {
                        return;
                    }
                    this.setCharacterData(false);
                    if (this.fCurrentNode != this.fRoot && !this.fInEntityRef && (this.fDOMFilter.getWhatToShow() & 0x1) != 0x0) {
                        short code = this.fDOMFilter.acceptNode(this.fCurrentNode);
                        switch (code) {
                            case 4: {
                                throw AbstractDOMParser.Abort.INSTANCE;
                            }
                            case 2: {
                                Node parent = this.fCurrentNode.getParentNode();
                                parent.removeChild(this.fCurrentNode);
                                this.fCurrentNode = parent;
                                return;
                            }
                            case 3: {
                                this.fFirstChunk = true;
                                Node parent = this.fCurrentNode.getParentNode();
                                NodeList ls = this.fCurrentNode.getChildNodes();
                                for (int length = ls.getLength(), i = 0; i < length; ++i) {
                                    parent.appendChild(ls.item(0));
                                }
                                parent.removeChild(this.fCurrentNode);
                                this.fCurrentNode = parent;
                                return;
                            }
                            default: {
                                break;
                            }
                        }
                    }
                    this.fCurrentNode = this.fCurrentNode.getParentNode();
                }
            }
            else {
                this.setCharacterData(false);
                this.fCurrentNode = this.fCurrentNode.getParentNode();
            }
        }
        else {
            if (augs != null) {
                ElementPSVI elementPSVI = (ElementPSVI)augs.getItem("ELEMENT_PSVI");
                if (elementPSVI != null) {
                    XSTypeDefinition type = elementPSVI.getMemberTypeDefinition();
                    if (type == null) {
                        type = elementPSVI.getTypeDefinition();
                    }
                    this.fDeferredDocumentImpl.setTypeInfo(this.fCurrentNodeIndex, type);
                }
            }
            this.fCurrentNodeIndex = this.fDeferredDocumentImpl.getParentNode(this.fCurrentNodeIndex, false);
        }
    }
    
    @Override
    public void startCDATA(Augmentations augs) throws XNIException {
        this.fInCDATASection = true;
        if (!this.fDeferNodeExpansion) {
            if (this.fFilterReject) {
                return;
            }
            if (this.fCreateCDATANodes) {
                this.setCharacterData(false);
            }
        }
    }
    
    @Override
    public void endCDATA(Augmentations augs) throws XNIException {
        this.fInCDATASection = false;
        if (!this.fDeferNodeExpansion) {
            if (this.fFilterReject) {
                return;
            }
            if (this.fCurrentCDATASection != null) {
                if (this.fDOMFilter != null && !this.fInEntityRef && (this.fDOMFilter.getWhatToShow() & 0x8) != 0x0) {
                    short code = this.fDOMFilter.acceptNode(this.fCurrentCDATASection);
                    switch (code) {
                        case 4: {
                            throw AbstractDOMParser.Abort.INSTANCE;
                        }
                        case 2:
                        case 3: {
                            Node parent = this.fCurrentNode.getParentNode();
                            parent.removeChild(this.fCurrentCDATASection);
                            this.fCurrentNode = parent;
                            return;
                        }
                        default: {
                            break;
                        }
                    }
                }
                this.fCurrentNode = this.fCurrentNode.getParentNode();
                this.fCurrentCDATASection = null;
            }
        }
        else if (this.fCurrentCDATASectionIndex != -1) {
            this.fCurrentNodeIndex = this.fDeferredDocumentImpl.getParentNode(this.fCurrentNodeIndex, false);
            this.fCurrentCDATASectionIndex = -1;
        }
    }
    
    @Override
    public void endDocument(Augmentations augs) throws XNIException {
        if (!this.fDeferNodeExpansion) {
            if (this.fDocumentImpl != null) {
                if (this.fLocator != null && this.fLocator.getEncoding() != null) {
                    this.fDocumentImpl.setInputEncoding(this.fLocator.getEncoding());
                }
                this.fDocumentImpl.setStrictErrorChecking(true);
            }
            this.fCurrentNode = null;
        }
        else {
            if (this.fLocator != null && this.fLocator.getEncoding() != null) {
                this.fDeferredDocumentImpl.setInputEncoding(this.fLocator.getEncoding());
            }
            this.fCurrentNodeIndex = -1;
        }
    }
    
    @Override
    public void endGeneralEntity(String name, Augmentations augs) throws XNIException {
        if (!this.fDeferNodeExpansion) {
            if (this.fFilterReject) {
                return;
            }
            this.setCharacterData(true);
            if (this.fDocumentType != null) {
                NamedNodeMap entities = this.fDocumentType.getEntities();
                this.fCurrentEntityDecl = (EntityImpl)entities.getNamedItem(name);
                if (this.fCurrentEntityDecl != null) {
                    if (this.fCurrentEntityDecl != null && this.fCurrentEntityDecl.getFirstChild() == null) {
                        this.fCurrentEntityDecl.setReadOnly(false, true);
                        for (Node child = this.fCurrentNode.getFirstChild(); child != null; child = child.getNextSibling()) {
                            Node copy = child.cloneNode(true);
                            this.fCurrentEntityDecl.appendChild(copy);
                        }
                        this.fCurrentEntityDecl.setReadOnly(true, true);
                    }
                    this.fCurrentEntityDecl = null;
                }
            }
            this.fInEntityRef = false;
            boolean removeEntityRef = false;
            if (this.fCreateEntityRefNodes) {
                if (this.fDocumentImpl != null) {
                    ((NodeImpl)this.fCurrentNode).setReadOnly(true, true);
                }
                if (this.fDOMFilter != null && (this.fDOMFilter.getWhatToShow() & 0x10) != 0x0) {
                    short code = this.fDOMFilter.acceptNode(this.fCurrentNode);
                    switch (code) {
                        case 4: {
                            throw AbstractDOMParser.Abort.INSTANCE;
                        }
                        case 2: {
                            Node parent = this.fCurrentNode.getParentNode();
                            parent.removeChild(this.fCurrentNode);
                            this.fCurrentNode = parent;
                            return;
                        }
                        case 3: {
                            this.fFirstChunk = true;
                            removeEntityRef = true;
                            break;
                        }
                        default: {
                            this.fCurrentNode = this.fCurrentNode.getParentNode();
                            break;
                        }
                    }
                }
                else {
                    this.fCurrentNode = this.fCurrentNode.getParentNode();
                }
            }
            if (!this.fCreateEntityRefNodes || removeEntityRef) {
                NodeList children = this.fCurrentNode.getChildNodes();
                Node parent = this.fCurrentNode.getParentNode();
                int length = children.getLength();
                if (length > 0) {
                    Node node2 = this.fCurrentNode.getPreviousSibling();
                    Node child2 = children.item(0);
                    if (node2 != null && node2.getNodeType() == 3 && child2.getNodeType() == 3) {
                        ((Text)node2).appendData(child2.getNodeValue());
                        this.fCurrentNode.removeChild(child2);
                    }
                    else {
                        node2 = parent.insertBefore(child2, this.fCurrentNode);
                        this.handleBaseURI(node2);
                    }
                    for (int i = 1; i < length; ++i) {
                        node2 = parent.insertBefore(children.item(0), this.fCurrentNode);
                        this.handleBaseURI(node2);
                    }
                }
                parent.removeChild(this.fCurrentNode);
                this.fCurrentNode = parent;
            }
        }
        else {
            if (this.fDocumentTypeIndex != -1) {
                for (int node = this.fDeferredDocumentImpl.getLastChild(this.fDocumentTypeIndex, false); node != -1; node = this.fDeferredDocumentImpl.getRealPrevSibling(node, false)) {
                    short nodeType = this.fDeferredDocumentImpl.getNodeType(node, false);
                    if (nodeType == 6) {
                        String nodeName = this.fDeferredDocumentImpl.getNodeName(node, false);
                        if (nodeName.equals(name)) {
                            this.fDeferredEntityDecl = node;
                            break;
                        }
                    }
                }
            }
            if (this.fDeferredEntityDecl != -1 && this.fDeferredDocumentImpl.getLastChild(this.fDeferredEntityDecl, false) == -1) {
                int prevIndex = -1;
                int cloneIndex;
                for (int childIndex = this.fDeferredDocumentImpl.getLastChild(this.fCurrentNodeIndex, false); childIndex != -1; prevIndex = cloneIndex, childIndex = this.fDeferredDocumentImpl.getRealPrevSibling(childIndex, false)) {
                    cloneIndex = this.fDeferredDocumentImpl.cloneNode(childIndex, true);
                    this.fDeferredDocumentImpl.insertBefore(this.fDeferredEntityDecl, cloneIndex, prevIndex);
                }
            }
            if (this.fCreateEntityRefNodes) {
                this.fCurrentNodeIndex = this.fDeferredDocumentImpl.getParentNode(this.fCurrentNodeIndex, false);
            }
            else {
                int childIndex2 = this.fDeferredDocumentImpl.getLastChild(this.fCurrentNodeIndex, false);
                int parentIndex = this.fDeferredDocumentImpl.getParentNode(this.fCurrentNodeIndex, false);
                int prevIndex2 = this.fCurrentNodeIndex;
                int lastChild = childIndex2;
                for (int sibling = -1; childIndex2 != -1; prevIndex2 = childIndex2, childIndex2 = sibling) {
                    this.handleBaseURI(childIndex2);
                    sibling = this.fDeferredDocumentImpl.getRealPrevSibling(childIndex2, false);
                    this.fDeferredDocumentImpl.insertBefore(parentIndex, childIndex2, prevIndex2);
                }
                if (lastChild != -1) {
                    this.fDeferredDocumentImpl.setAsLastChild(parentIndex, lastChild);
                }
                else {
                    int sibling = this.fDeferredDocumentImpl.getRealPrevSibling(prevIndex2, false);
                    this.fDeferredDocumentImpl.setAsLastChild(parentIndex, sibling);
                }
                this.fCurrentNodeIndex = parentIndex;
            }
            this.fDeferredEntityDecl = -1;
        }
    }
    
    protected final void handleBaseURI(Node node) {
        if (this.fDocumentImpl != null) {
            String baseURI = null;
            short nodeType = node.getNodeType();
            if (nodeType == 1) {
                if (this.fNamespaceAware) {
                    if (((Element)node).getAttributeNodeNS("http://www.w3.org/XML/1998/namespace", "base") != null) {
                        return;
                    }
                }
                else if (((Element)node).getAttributeNode("xml:base") != null) {
                    return;
                }
                baseURI = ((EntityReferenceImpl)this.fCurrentNode).getBaseURI();
                if (baseURI != null && !baseURI.equals(this.fDocumentImpl.getDocumentURI())) {
                    if (this.fNamespaceAware) {
                        ((Element)node).setAttributeNS("http://www.w3.org/XML/1998/namespace", "xml:base", baseURI);
                    }
                    else {
                        ((Element)node).setAttribute("xml:base", baseURI);
                    }
                }
            }
            else if (nodeType == 7) {
                baseURI = ((EntityReferenceImpl)this.fCurrentNode).getBaseURI();
                if (baseURI != null && this.fErrorHandler != null) {
                    DOMErrorImpl error = new DOMErrorImpl();
                    error.fType = "pi-base-uri-not-preserved";
                    error.fRelatedData = baseURI;
                    error.fSeverity = 1;
                    this.fErrorHandler.getErrorHandler().handleError(error);
                }
            }
        }
    }
    
    protected final void handleBaseURI(int node) {
        short nodeType = this.fDeferredDocumentImpl.getNodeType(node, false);
        if (nodeType == 1) {
            String baseURI = this.fDeferredDocumentImpl.getNodeValueString(this.fCurrentNodeIndex, false);
            if (baseURI == null) {
                baseURI = this.fDeferredDocumentImpl.getDeferredEntityBaseURI(this.fDeferredEntityDecl);
            }
            if (baseURI != null && !baseURI.equals(this.fDeferredDocumentImpl.getDocumentURI())) {
                this.fDeferredDocumentImpl.setDeferredAttribute(node, "xml:base", "http://www.w3.org/XML/1998/namespace", baseURI, true, false, null);
            }
        }
        else if (nodeType == 7) {
            String baseURI = this.fDeferredDocumentImpl.getNodeValueString(this.fCurrentNodeIndex, false);
            if (baseURI == null) {
                baseURI = this.fDeferredDocumentImpl.getDeferredEntityBaseURI(this.fDeferredEntityDecl);
            }
            if (baseURI != null && this.fErrorHandler != null) {
                DOMErrorImpl error = new DOMErrorImpl();
                error.fType = "pi-base-uri-not-preserved";
                error.fRelatedData = baseURI;
                error.fSeverity = 1;
                this.fErrorHandler.getErrorHandler().handleError(error);
            }
        }
    }
    
    @Override
    public void startDTD(XMLLocator locator, Augmentations augs) throws XNIException {
        this.fInDTD = true;
        if (locator != null) {
            this.fBaseURIStack.push(locator.getBaseSystemId());
        }
        if (this.fDeferNodeExpansion || this.fDocumentImpl != null) {
            this.fInternalSubset = new StringBuilder(1024);
        }
    }
    
    @Override
    public void endDTD(Augmentations augs) throws XNIException {
        this.fInDTD = false;
        if (!this.fBaseURIStack.isEmpty()) {
            this.fBaseURIStack.pop();
        }
        String internalSubset = (this.fInternalSubset != null && this.fInternalSubset.length() > 0) ? this.fInternalSubset.toString() : null;
        if (this.fDeferNodeExpansion) {
            if (internalSubset != null) {
                this.fDeferredDocumentImpl.setInternalSubset(this.fDocumentTypeIndex, internalSubset);
            }
        }
        else if (this.fDocumentImpl != null && internalSubset != null) {
            ((DocumentTypeImpl)this.fDocumentType).setInternalSubset(internalSubset);
        }
    }
    
    @Override
    public void startConditional(short type, Augmentations augs) throws XNIException {
    }
    
    @Override
    public void endConditional(Augmentations augs) throws XNIException {
    }
    
    @Override
    public void startExternalSubset(XMLResourceIdentifier identifier, Augmentations augs) throws XNIException {
        this.fBaseURIStack.push(identifier.getBaseSystemId());
        this.fInDTDExternalSubset = true;
    }
    
    @Override
    public void endExternalSubset(Augmentations augs) throws XNIException {
        this.fInDTDExternalSubset = false;
        this.fBaseURIStack.pop();
    }
    
    @Override
    public void internalEntityDecl(String name, XMLString text, XMLString nonNormalizedText, Augmentations augs) throws XNIException {
        if (this.fInternalSubset != null && !this.fInDTDExternalSubset) {
            this.fInternalSubset.append("<!ENTITY ");
            if (name.startsWith("%")) {
                this.fInternalSubset.append("% ");
                this.fInternalSubset.append(name.substring(1));
            }
            else {
                this.fInternalSubset.append(name);
            }
            this.fInternalSubset.append(' ');
            String value = nonNormalizedText.toString();
            boolean singleQuote = value.indexOf(39) == -1;
            this.fInternalSubset.append(singleQuote ? '\'' : '\"');
            this.fInternalSubset.append(value);
            this.fInternalSubset.append(singleQuote ? '\'' : '\"');
            this.fInternalSubset.append(">\n");
        }
        if (name.startsWith("%")) {
            return;
        }
        if (this.fDocumentType != null) {
            NamedNodeMap entities = this.fDocumentType.getEntities();
            EntityImpl entity = (EntityImpl)entities.getNamedItem(name);
            if (entity == null) {
                entity = (EntityImpl)this.fDocumentImpl.createEntity(name);
                entity.setBaseURI((String)this.fBaseURIStack.peek());
                entities.setNamedItem(entity);
            }
        }
        if (this.fDocumentTypeIndex != -1) {
            boolean found = false;
            for (int node = this.fDeferredDocumentImpl.getLastChild(this.fDocumentTypeIndex, false); node != -1; node = this.fDeferredDocumentImpl.getRealPrevSibling(node, false)) {
                short nodeType = this.fDeferredDocumentImpl.getNodeType(node, false);
                if (nodeType == 6) {
                    String nodeName = this.fDeferredDocumentImpl.getNodeName(node, false);
                    if (nodeName.equals(name)) {
                        found = true;
                        break;
                    }
                }
            }
            if (!found) {
                int entityIndex = this.fDeferredDocumentImpl.createDeferredEntity(name, null, null, null, (String)this.fBaseURIStack.peek());
                this.fDeferredDocumentImpl.appendChild(this.fDocumentTypeIndex, entityIndex);
            }
        }
        return;
    }
    
    @Override
    public void externalEntityDecl(String name, XMLResourceIdentifier identifier, Augmentations augs) throws XNIException {
        String publicId = identifier.getPublicId();
        String literalSystemId = identifier.getLiteralSystemId();
        if (this.fInternalSubset != null && !this.fInDTDExternalSubset) {
            this.fInternalSubset.append("<!ENTITY ");
            if (name.startsWith("%")) {
                this.fInternalSubset.append("% ");
                this.fInternalSubset.append(name.substring(1));
            }
            else {
                this.fInternalSubset.append(name);
            }
            this.fInternalSubset.append(' ');
            if (publicId != null) {
                this.fInternalSubset.append("PUBLIC '");
                this.fInternalSubset.append(publicId);
                this.fInternalSubset.append("' '");
            }
            else {
                this.fInternalSubset.append("SYSTEM '");
            }
            this.fInternalSubset.append(literalSystemId);
            this.fInternalSubset.append("'>\n");
        }
        if (name.startsWith("%")) {
            return;
        }
        if (this.fDocumentType != null) {
            NamedNodeMap entities = this.fDocumentType.getEntities();
            EntityImpl entity = (EntityImpl)entities.getNamedItem(name);
            if (entity == null) {
                entity = (EntityImpl)this.fDocumentImpl.createEntity(name);
                entity.setPublicId(publicId);
                entity.setSystemId(literalSystemId);
                entity.setBaseURI(identifier.getBaseSystemId());
                entities.setNamedItem(entity);
            }
        }
        if (this.fDocumentTypeIndex != -1) {
            boolean found = false;
            for (int nodeIndex = this.fDeferredDocumentImpl.getLastChild(this.fDocumentTypeIndex, false); nodeIndex != -1; nodeIndex = this.fDeferredDocumentImpl.getRealPrevSibling(nodeIndex, false)) {
                short nodeType = this.fDeferredDocumentImpl.getNodeType(nodeIndex, false);
                if (nodeType == 6) {
                    String nodeName = this.fDeferredDocumentImpl.getNodeName(nodeIndex, false);
                    if (nodeName.equals(name)) {
                        found = true;
                        break;
                    }
                }
            }
            if (!found) {
                int entityIndex = this.fDeferredDocumentImpl.createDeferredEntity(name, publicId, literalSystemId, null, identifier.getBaseSystemId());
                this.fDeferredDocumentImpl.appendChild(this.fDocumentTypeIndex, entityIndex);
            }
        }
        return;
    }
    
    @Override
    public void startParameterEntity(String name, XMLResourceIdentifier identifier, String encoding, Augmentations augs) throws XNIException {
        if (augs != null && this.fInternalSubset != null && !this.fInDTDExternalSubset && Boolean.TRUE.equals(augs.getItem("ENTITY_SKIPPED"))) {
            this.fInternalSubset.append(name).append(";\n");
        }
        this.fBaseURIStack.push(identifier.getExpandedSystemId());
    }
    
    @Override
    public void endParameterEntity(String name, Augmentations augs) throws XNIException {
        this.fBaseURIStack.pop();
    }
    
    @Override
    public void unparsedEntityDecl(String name, XMLResourceIdentifier identifier, String notation, Augmentations augs) throws XNIException {
        String publicId = identifier.getPublicId();
        String literalSystemId = identifier.getLiteralSystemId();
        if (this.fInternalSubset != null && !this.fInDTDExternalSubset) {
            this.fInternalSubset.append("<!ENTITY ");
            this.fInternalSubset.append(name);
            this.fInternalSubset.append(' ');
            if (publicId != null) {
                this.fInternalSubset.append("PUBLIC '");
                this.fInternalSubset.append(publicId);
                if (literalSystemId != null) {
                    this.fInternalSubset.append("' '");
                    this.fInternalSubset.append(literalSystemId);
                }
            }
            else {
                this.fInternalSubset.append("SYSTEM '");
                this.fInternalSubset.append(literalSystemId);
            }
            this.fInternalSubset.append("' NDATA ");
            this.fInternalSubset.append(notation);
            this.fInternalSubset.append(">\n");
        }
        if (this.fDocumentType != null) {
            NamedNodeMap entities = this.fDocumentType.getEntities();
            EntityImpl entity = (EntityImpl)entities.getNamedItem(name);
            if (entity == null) {
                entity = (EntityImpl)this.fDocumentImpl.createEntity(name);
                entity.setPublicId(publicId);
                entity.setSystemId(literalSystemId);
                entity.setNotationName(notation);
                entity.setBaseURI(identifier.getBaseSystemId());
                entities.setNamedItem(entity);
            }
        }
        if (this.fDocumentTypeIndex != -1) {
            boolean found = false;
            for (int nodeIndex = this.fDeferredDocumentImpl.getLastChild(this.fDocumentTypeIndex, false); nodeIndex != -1; nodeIndex = this.fDeferredDocumentImpl.getRealPrevSibling(nodeIndex, false)) {
                short nodeType = this.fDeferredDocumentImpl.getNodeType(nodeIndex, false);
                if (nodeType == 6) {
                    String nodeName = this.fDeferredDocumentImpl.getNodeName(nodeIndex, false);
                    if (nodeName.equals(name)) {
                        found = true;
                        break;
                    }
                }
            }
            if (!found) {
                int entityIndex = this.fDeferredDocumentImpl.createDeferredEntity(name, publicId, literalSystemId, notation, identifier.getBaseSystemId());
                this.fDeferredDocumentImpl.appendChild(this.fDocumentTypeIndex, entityIndex);
            }
        }
    }
    
    @Override
    public void notationDecl(String name, XMLResourceIdentifier identifier, Augmentations augs) throws XNIException {
        String publicId = identifier.getPublicId();
        String literalSystemId = identifier.getLiteralSystemId();
        if (this.fInternalSubset != null && !this.fInDTDExternalSubset) {
            this.fInternalSubset.append("<!NOTATION ");
            this.fInternalSubset.append(name);
            if (publicId != null) {
                this.fInternalSubset.append(" PUBLIC '");
                this.fInternalSubset.append(publicId);
                if (literalSystemId != null) {
                    this.fInternalSubset.append("' '");
                    this.fInternalSubset.append(literalSystemId);
                }
            }
            else {
                this.fInternalSubset.append(" SYSTEM '");
                this.fInternalSubset.append(literalSystemId);
            }
            this.fInternalSubset.append("'>\n");
        }
        if (this.fDocumentImpl != null && this.fDocumentType != null) {
            NamedNodeMap notations = this.fDocumentType.getNotations();
            if (notations.getNamedItem(name) == null) {
                NotationImpl notation = (NotationImpl)this.fDocumentImpl.createNotation(name);
                notation.setPublicId(publicId);
                notation.setSystemId(literalSystemId);
                notation.setBaseURI(identifier.getBaseSystemId());
                notations.setNamedItem(notation);
            }
        }
        if (this.fDocumentTypeIndex != -1) {
            boolean found = false;
            for (int nodeIndex = this.fDeferredDocumentImpl.getLastChild(this.fDocumentTypeIndex, false); nodeIndex != -1; nodeIndex = this.fDeferredDocumentImpl.getPrevSibling(nodeIndex, false)) {
                short nodeType = this.fDeferredDocumentImpl.getNodeType(nodeIndex, false);
                if (nodeType == 12) {
                    String nodeName = this.fDeferredDocumentImpl.getNodeName(nodeIndex, false);
                    if (nodeName.equals(name)) {
                        found = true;
                        break;
                    }
                }
            }
            if (!found) {
                int notationIndex = this.fDeferredDocumentImpl.createDeferredNotation(name, publicId, literalSystemId, identifier.getBaseSystemId());
                this.fDeferredDocumentImpl.appendChild(this.fDocumentTypeIndex, notationIndex);
            }
        }
    }
    
    @Override
    public void ignoredCharacters(XMLString text, Augmentations augs) throws XNIException {
    }
    
    @Override
    public void elementDecl(String name, String contentModel, Augmentations augs) throws XNIException {
        if (this.fInternalSubset != null && !this.fInDTDExternalSubset) {
            this.fInternalSubset.append("<!ELEMENT ");
            this.fInternalSubset.append(name);
            this.fInternalSubset.append(' ');
            this.fInternalSubset.append(contentModel);
            this.fInternalSubset.append(">\n");
        }
    }
    
    @Override
    public void attributeDecl(String elementName, String attributeName, String type, String[] enumeration, String defaultType, XMLString defaultValue, XMLString nonNormalizedDefaultValue, Augmentations augs) throws XNIException {
        if (this.fInternalSubset != null && !this.fInDTDExternalSubset) {
            this.fInternalSubset.append("<!ATTLIST ");
            this.fInternalSubset.append(elementName);
            this.fInternalSubset.append(' ');
            this.fInternalSubset.append(attributeName);
            this.fInternalSubset.append(' ');
            if (type.equals("ENUMERATION")) {
                this.fInternalSubset.append('(');
                for (int i = 0; i < enumeration.length; ++i) {
                    if (i > 0) {
                        this.fInternalSubset.append('|');
                    }
                    this.fInternalSubset.append(enumeration[i]);
                }
                this.fInternalSubset.append(')');
            }
            else {
                this.fInternalSubset.append(type);
            }
            if (defaultType != null) {
                this.fInternalSubset.append(' ');
                this.fInternalSubset.append(defaultType);
            }
            if (defaultValue != null) {
                this.fInternalSubset.append(" '");
                for (int i = 0; i < defaultValue.length; ++i) {
                    char c = defaultValue.ch[defaultValue.offset + i];
                    if (c == '\'') {
                        this.fInternalSubset.append("&apos;");
                    }
                    else {
                        this.fInternalSubset.append(c);
                    }
                }
                this.fInternalSubset.append('\'');
            }
            this.fInternalSubset.append(">\n");
        }
        if (this.fDeferredDocumentImpl != null) {
            if (defaultValue != null) {
                int elementDefIndex = this.fDeferredDocumentImpl.lookupElementDefinition(elementName);
                if (elementDefIndex == -1) {
                    elementDefIndex = this.fDeferredDocumentImpl.createDeferredElementDefinition(elementName);
                    this.fDeferredDocumentImpl.appendChild(this.fDocumentTypeIndex, elementDefIndex);
                }
                boolean nsEnabled2 = this.fNamespaceAware;
                String namespaceURI2 = null;
                if (nsEnabled2) {
                    if (attributeName.startsWith("xmlns:") || attributeName.equals("xmlns")) {
                        namespaceURI2 = NamespaceContext.XMLNS_URI;
                    }
                    else if (attributeName.startsWith("xml:")) {
                        namespaceURI2 = NamespaceContext.XML_URI;
                    }
                }
                int attrIndex = this.fDeferredDocumentImpl.createDeferredAttribute(attributeName, namespaceURI2, defaultValue.toString(), false);
                if ("ID".equals(type)) {
                    this.fDeferredDocumentImpl.setIdAttribute(attrIndex);
                }
                this.fDeferredDocumentImpl.appendChild(elementDefIndex, attrIndex);
            }
        }
        else if (this.fDocumentImpl != null && defaultValue != null) {
            NamedNodeMap elements = ((DocumentTypeImpl)this.fDocumentType).getElements();
            ElementDefinitionImpl elementDef = (ElementDefinitionImpl)elements.getNamedItem(elementName);
            if (elementDef == null) {
                elementDef = this.fDocumentImpl.createElementDefinition(elementName);
                ((DocumentTypeImpl)this.fDocumentType).getElements().setNamedItem(elementDef);
            }
            boolean nsEnabled = this.fNamespaceAware;
            AttrImpl attr;
            if (nsEnabled) {
                String namespaceURI = null;
                if (attributeName.startsWith("xmlns:") || attributeName.equals("xmlns")) {
                    namespaceURI = NamespaceContext.XMLNS_URI;
                }
                else if (attributeName.startsWith("xml:")) {
                    namespaceURI = NamespaceContext.XML_URI;
                }
                attr = (AttrImpl)this.fDocumentImpl.createAttributeNS(namespaceURI, attributeName);
            }
            else {
                attr = (AttrImpl)this.fDocumentImpl.createAttribute(attributeName);
            }
            attr.setValue(defaultValue.toString());
            attr.setSpecified(false);
            attr.setIdAttribute("ID".equals(type));
            if (nsEnabled) {
                elementDef.getAttributes().setNamedItemNS(attr);
            }
            else {
                elementDef.getAttributes().setNamedItem(attr);
            }
        }
    }
    
    @Override
    public void startAttlist(String elementName, Augmentations augs) throws XNIException {
    }
    
    @Override
    public void endAttlist(Augmentations augs) throws XNIException {
    }
    
    protected Element createElementNode(QName element) {
        Element el = null;
        if (this.fNamespaceAware) {
            if (this.fDocumentImpl != null) {
                el = this.fDocumentImpl.createElementNS(element.uri, element.rawname, element.localpart);
            }
            else {
                el = this.fDocument.createElementNS(element.uri, element.rawname);
            }
        }
        else {
            el = this.fDocument.createElement(element.rawname);
        }
        return el;
    }
    
    protected Attr createAttrNode(QName attrQName) {
        Attr attr = null;
        if (this.fNamespaceAware) {
            if (this.fDocumentImpl != null) {
                attr = this.fDocumentImpl.createAttributeNS(attrQName.uri, attrQName.rawname, attrQName.localpart);
            }
            else {
                attr = this.fDocument.createAttributeNS(attrQName.uri, attrQName.rawname);
            }
        }
        else {
            attr = this.fDocument.createAttribute(attrQName.rawname);
        }
        return attr;
    }
    
    protected void setCharacterData(boolean sawChars) {
        this.fFirstChunk = sawChars;
        Node child = this.fCurrentNode.getLastChild();
        if (child != null) {
            if (this.fStringBuilder.length() > 0) {
                if (child.getNodeType() == 3) {
                    if (this.fDocumentImpl != null) {
                        ((TextImpl)child).replaceData(this.fStringBuilder.toString());
                    }
                    else {
                        ((Text)child).setData(this.fStringBuilder.toString());
                    }
                }
                this.fStringBuilder.setLength(0);
            }
            if (this.fDOMFilter != null && !this.fInEntityRef && child.getNodeType() == 3 && (this.fDOMFilter.getWhatToShow() & 0x4) != 0x0) {
                short code = this.fDOMFilter.acceptNode(child);
                switch (code) {
                    case 4: {
                        throw AbstractDOMParser.Abort.INSTANCE;
                    }
                    case 2:
                    case 3: {
                        this.fCurrentNode.removeChild(child);
                        return;
                    }
                    default: {
                        break;
                    }
                }
            }
        }
    }
    
    public void abort() {
        throw AbstractDOMParser.Abort.INSTANCE;
    }
    
    static {
        RECOGNIZED_FEATURES = new String[] { "http://xml.org/sax/features/namespaces", "http://apache.org/xml/features/dom/create-entity-ref-nodes", "http://apache.org/xml/features/include-comments", "http://apache.org/xml/features/create-cdata-nodes", "http://apache.org/xml/features/dom/include-ignorable-whitespace", "http://apache.org/xml/features/dom/defer-node-expansion" };
        RECOGNIZED_PROPERTIES = new String[] { "http://apache.org/xml/properties/dom/document-class-name", "http://apache.org/xml/properties/dom/current-element-node" };
    }
    
    static final class Abort extends RuntimeException
    {
        private static final long serialVersionUID = 1687848994976808490L;
        static final AbstractDOMParser.Abort INSTANCE;
        
        private Abort() {
        }
        
        @Override
        public Throwable fillInStackTrace() {
            return this;
        }
        
        static {
            INSTANCE = new AbstractDOMParser.Abort();
        }
    }
}
